import pytest
from .api_versions import ROOT_ALL_VERSIONS


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_default_workflows(rest_client, default_workflow_identifiers, root):
    for identifier in default_workflow_identifiers:
        response = rest_client.get(f"{root}/workflow/{identifier}")
        data = response.json()
        assert response.status_code == 200, data


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_default_icons(rest_client, default_icon_identifiers, root):
    for identifier in default_icon_identifiers:
        response = rest_client.get(f"{root}/icon/{identifier}")
        data = response.json()
        assert response.status_code == 200, data


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_default_tasks(rest_client, default_task_identifiers, root):
    for identifier in default_task_identifiers:
        response = rest_client.get(f"{root}/task/{identifier}")
        data = response.json()
        assert response.status_code == 200, data



from ..app import backend


def test_get_routes():
    routes = {(1, 0, 0): None, (1, 0, 1): None, (2, 0, 0): None}

    all_routes = backend.get_routes("test", routes)
    expected = {
        (1, 0, 0, 0): backend.Route(
            router=None, prefix="/api/v1_0_0", tag="v1.0.0", versioned=True
        ),
        (1, 0, 1, 0): backend.Route(
            router=None, prefix="/api/v1_0_1", tag="v1.0.1", versioned=True
        ),
        (2, 0, 0, 0): backend.Route(
            router=None, prefix="/api/v2_0_0", tag="v2.0.0", versioned=True
        ),
        (1, 0, 1, 1): backend.Route(
            router=None, prefix="/api/v1", tag="v1", versioned=True
        ),
        (2, 0, 0, 1): backend.Route(
            router=None, prefix="/api/v2", tag="v2", versioned=True
        ),
        (2, 0, 0, 2): backend.Route(
            router=None, prefix="/api", tag="test", versioned=False
        ),
    }
    assert all_routes == expected

    all_routes = backend.get_routes("test", routes, suffix="testprefix")
    expected = {
        (1, 0, 0, 0): backend.Route(
            router=None, prefix="/api/v1_0_0/testprefix", tag="v1.0.0", versioned=True
        ),
        (1, 0, 1, 0): backend.Route(
            router=None, prefix="/api/v1_0_1/testprefix", tag="v1.0.1", versioned=True
        ),
        (2, 0, 0, 0): backend.Route(
            router=None, prefix="/api/v2_0_0/testprefix", tag="v2.0.0", versioned=True
        ),
        (1, 0, 1, 1): backend.Route(
            router=None, prefix="/api/v1/testprefix", tag="v1", versioned=True
        ),
        (2, 0, 0, 1): backend.Route(
            router=None, prefix="/api/v2/testprefix", tag="v2", versioned=True
        ),
        (2, 0, 0, 2): backend.Route(
            router=None, prefix="/api/testprefix", tag="test", versioned=False
        ),
    }
    assert all_routes == expected



from typing import Optional, Dict
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware


def enable_cors(app: FastAPI) -> None:
    """Enable Cross-Origin Resource Sharing"""
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )


def get_cors_options(app: FastAPI) -> Optional[Dict]:
    for mw in app.user_middleware:
        if mw.cls is CORSMiddleware:
            return mw.options



from dataclasses import dataclass
from typing import Dict, Tuple, List, Set, Union, Callable

from fastapi import APIRouter
from fastapi import FastAPI
from starlette.types import ASGIApp

from . import BACKEND_PREFIX

AppGenerator = Callable[[], ASGIApp]
RouterType = Union[APIRouter, AppGenerator]


@dataclass
class Route:
    router: RouterType
    prefix: str
    tag: str
    versioned: bool


def get_routes(
    tag: str, routers: Dict[Tuple[int], RouterType], suffix: str = ""
) -> Dict[Tuple[int], Route]:
    """Generate routes with versioned paths for all strict and major versions.
    In addition add a route with non-versioned path for the latest version."""
    routes = dict()

    major_routes = dict()
    major_versions = dict()
    for full_version, router in reversed(sorted(routers.items())):
        assert len(full_version) == 3, full_version
        major, minor, patch = full_version
        route_key = major, minor, patch, 0
        path_version = "v" + "_".join(map(str, full_version))
        route_tag = "v" + ".".join(map(str, full_version))
        if suffix:
            prefix = f"{BACKEND_PREFIX}/{path_version}/{suffix}"
        else:
            prefix = f"{BACKEND_PREFIX}/{path_version}"
        routes[route_key] = Route(
            router=router, prefix=prefix, tag=route_tag, versioned=True
        )
        if full_version > major_versions.get(major, (0, 0, 0)):
            major_versions[major] = full_version
            major_routes[major] = router

    for major, router in major_routes.items():
        path_version = route_tag = f"v{major}"
        full_version = major_versions[major]
        _, minor, patch = full_version
        route_key = major, minor, patch, 1
        if suffix:
            prefix = f"{BACKEND_PREFIX}/{path_version}/{suffix}"
        else:
            prefix = f"{BACKEND_PREFIX}/{path_version}"
        routes[route_key] = Route(
            router=router, prefix=prefix, tag=route_tag, versioned=True
        )

    last_major_version = sorted(major_routes)[-1]
    router = major_routes[last_major_version]
    major, minor, patch = major_versions[last_major_version]
    route_key = major, minor, patch, 2
    if suffix:
        prefix = f"{BACKEND_PREFIX}/{suffix}"
    else:
        prefix = f"{BACKEND_PREFIX}"
    routes[route_key] = Route(router=router, prefix=prefix, tag=tag, versioned=False)
    return routes


def assert_route_versions(*all_routes: Dict[Tuple[int], RouterType]) -> None:
    versions = {tuple(sorted(routes)) for routes in all_routes}
    assert len(versions) == 1, "Not all routes have the same versions"


def extract_version_tags(all_routes: List[Dict[Tuple[int], Route]]) -> Set[str]:
    """Extract all version tags"""
    tags = set()
    for routes in all_routes:
        for route in routes.values():
            if route.versioned:
                tags.add(route.tag)
    return tags


def extract_latest_version(all_routes: List[Dict[Tuple[int], Route]]) -> Tuple[int]:
    """Extract the latest version"""
    return max(sorted(routes)[-1][:3] for routes in all_routes)


def add_routes(
    app: FastAPI,
    all_routes: List[Dict[Tuple[int], Route]],
    no_older_versions: bool = False,
) -> None:
    """Add routes to a fastapi app"""
    route_keys = set()
    for keys in all_routes:
        route_keys |= set(keys)

    for route_key in reversed(sorted(route_keys)):
        for routes in all_routes:
            route = routes.get(route_key)
            if route is None:
                continue
            if no_older_versions and route.versioned:
                continue
            if isinstance(route.router, APIRouter):
                app.include_router(
                    route.router,
                    prefix=route.prefix,
                    tags=[route.tag],
                )
            elif isinstance(route.router, Callable):
                subapp = route.router()
                app.mount(route.prefix, subapp)
            else:
                raise TypeError(str(type(route)))



from typing import Optional, Dict, List
from ewoksjob.client import discover_all_tasks
from ewoksjob.client.local import discover_all_tasks as discover_all_tasks_local
from ewoksjob.client import discover_tasks_from_modules
from ewoksjob.client.local import (
    discover_tasks_from_modules as discover_tasks_from_modules_local,
)

from ...config import EwoksSettings


def discover_tasks(
    settings: EwoksSettings,
    modules: Optional[List[str]] = None,
    reload: Optional[bool] = None,
    worker_options: Optional[Dict] = None,
) -> List[Dict[str, str]]:
    if worker_options is None:
        kwargs = dict()
    else:
        kwargs = dict(worker_options)

    # Task discovery: position arguments
    if modules:
        kwargs["args"] = modules
    # Task discovery: named arguments
    kwargs["kwargs"] = dict()
    if reload is not None:
        kwargs["kwargs"]["reload"] = reload

    if settings.celery is None:
        if modules:
            future = discover_tasks_from_modules_local(**kwargs)
        else:
            future = discover_all_tasks_local(**kwargs)
        tasks = future.result()
    else:
        if modules:
            future = discover_tasks_from_modules(**kwargs)
        else:
            future = discover_all_tasks(**kwargs)
        tasks = future.get()

    for task in tasks:
        _set_default_task_properties(task)
    return tasks


def _set_default_task_properties(task: dict) -> None:
    if not task.get("icon"):
        task["icon"] = "default.png"
    if not task.get("label"):
        task_identifier = task.get("task_identifier")
        if task_identifier:
            task["label"] = task_identifier.split(".")[-1]



import logging
from typing import List, Dict
from typing_extensions import Annotated

from fastapi import APIRouter
from fastapi import Path
from fastapi import Body
from fastapi.responses import JSONResponse
from fastapi import status
from pydantic import ValidationError


from ...backends import json_backend
from ...config import EwoksSettingsType
from ..common import models as common_models
from . import models
from . import discovery

logger = logging.getLogger(__name__)

router = APIRouter()


@router.get(
    "/task/{identifier}",
    summary="Get ewoks task description",
    response_model=models.EwoksTaskDescription,
    response_model_exclude_none=True,
    response_description="Ewoks task description",
    status_code=200,
    responses={
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to read task",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_404_NOT_FOUND: {
            "description": "Task not found",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def get_task(
    identifier: Annotated[
        str,
        Path(
            title="Task identifier",
            description="Unique identifier in the task database",
        ),
    ],
    settings: EwoksSettingsType,
) -> json_backend.ResourceContentType:
    try:
        return json_backend.load_resource(
            settings.resource_directory / "tasks", identifier
        )
    except PermissionError:
        return JSONResponse(
            {
                "message": f"No permission to read task '{identifier}'.",
                "type": "task",
                "identifier": identifier,
            },
            status_code=status.HTTP_403_FORBIDDEN,
        )
    except FileNotFoundError:
        return JSONResponse(
            {
                "message": f"Task '{identifier}' is not found.",
                "type": "task",
                "identifier": identifier,
            },
            status_code=status.HTTP_404_NOT_FOUND,
        )


@router.get(
    "/tasks/descriptions",
    summary="Get all ewoks task descriptions",
    response_model=models.EwoksTaskDescriptions,
    response_description="Ewoks task descriptions",
    status_code=200,
)
def get_tasks(
    settings: EwoksSettingsType,
) -> Dict[str, List[models.EwoksTaskDescription]]:
    tasks = list(json_backend.resources(settings.resource_directory / "tasks"))

    valid_tasks = []
    for task in tasks:
        try:
            valid_task = models.EwoksTaskDescription(**task)
            valid_tasks.append(valid_task)
        except ValidationError as e:
            logger.warning(f"Invalid task description: {e}")

    return {"items": valid_tasks}


@router.get(
    "/tasks",
    summary="Get all ewoks task identifiers",
    response_model=models.EwoksTaskIdentifiers,
    response_description="Ewoks task identifiers",
    status_code=200,
)
def get_task_identifiers(settings: EwoksSettingsType) -> Dict[str, List[str]]:
    task_descriptions = get_tasks(settings)
    identifiers = [task.task_identifier for task in task_descriptions["items"]]
    return {"identifiers": identifiers}


@router.put(
    "/task/{identifier}",
    summary="Update ewoks task description",
    response_model=models.EwoksTaskDescription,
    response_model_exclude_none=True,
    response_description="Ewoks task description",
    status_code=200,
    responses={
        status.HTTP_400_BAD_REQUEST: {
            "description": "Wrong task identifier",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_404_NOT_FOUND: {
            "description": "Task not found",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to edit task",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def update_task(
    identifier: Annotated[
        str,
        Path(
            title="Task identifier",
            description="Unique identifier in the task database",
        ),
    ],
    task: Annotated[models.EwoksTaskDescription, Body(title="Ewoks task description")],
    settings: EwoksSettingsType,
) -> models.EwoksTaskDescription:
    ridentifier = task.task_identifier
    if identifier != ridentifier:
        return JSONResponse(
            {
                "message": f"Resource identifier '{identifier}' is not equal to '{ridentifier}'.",
                "type": "task",
                "identifier": identifier,
            },
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    exists = json_backend.resource_exists(
        settings.resource_directory / "tasks", identifier
    )
    if not exists:
        return JSONResponse(
            {
                "message": f"Task '{identifier}' is not found.",
                "type": "task",
                "identifier": identifier,
            },
            status_code=status.HTTP_404_NOT_FOUND,
        )

    try:
        json_backend.save_resource(
            settings.resource_directory / "tasks", identifier, task.model_dump()
        )
    except PermissionError:
        return JSONResponse(
            {
                "message": f"No permission to edit task '{identifier}'.",
                "type": "task",
                "identifier": identifier,
            },
            status_code=status.HTTP_403_FORBIDDEN,
        )

    return task


@router.post(
    "/tasks",
    summary="Create ewoks task description",
    response_model=models.EwoksTaskDescription,
    response_model_exclude_none=True,
    response_description="Ewoks task description",
    status_code=200,
    responses={
        status.HTTP_409_CONFLICT: {
            "description": "Task already exists",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to create task",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def create_task(
    task: Annotated[models.EwoksTaskDescription, Body(title="Ewoks task description")],
    settings: EwoksSettingsType,
) -> models.EwoksTaskDescription:
    ridentifier = task.task_identifier

    exists = json_backend.resource_exists(
        settings.resource_directory / "tasks", ridentifier
    )
    if exists:
        return JSONResponse(
            {
                "message": f"Task '{ridentifier}' already exists.",
                "type": "task",
                "identifier": ridentifier,
            },
            status_code=status.HTTP_409_CONFLICT,
        )

    try:
        json_backend.save_resource(
            settings.resource_directory / "tasks", ridentifier, task.model_dump()
        )
    except PermissionError:
        return JSONResponse(
            {
                "message": f"No permission to create task '{ridentifier}'.",
                "type": "task",
                "identifier": ridentifier,
            },
            status_code=status.HTTP_403_FORBIDDEN,
        )

    return task


@router.post(
    "/tasks/discover",
    summary="Create ewoks task descriptions from a worker environment",
    response_model=models.EwoksTaskIdentifiers,
    response_description="Discovered ewoks task identifiers",
    status_code=200,
    responses={
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to create or edit task",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_404_NOT_FOUND: {
            "description": "Module not found",
            "model": common_models.ResourceError,
        },
    },
)
def discover_tasks(
    settings: EwoksSettingsType,
    options: Annotated[
        models.EwoksTaskDiscovery, Body(title="Ewoks task discovery options")
    ] = None,
) -> Dict[str, List[str]]:
    if options:
        discover_options = options.model_dump()
    else:
        discover_options = dict()
    try:
        tasks = discovery.discover_tasks(settings, **discover_options)
    except ModuleNotFoundError as e:
        return JSONResponse(
            {
                "message": str(e),
                "type": "task",
            },
            status_code=status.HTTP_404_NOT_FOUND,
        )

    for task in tasks:
        ridentifier = task["task_identifier"]
        try:
            json_backend.save_resource(
                settings.resource_directory / "tasks", ridentifier, task
            )
        except PermissionError:
            return JSONResponse(
                {
                    "message": f"No permission to create task '{ridentifier}'.",
                    "type": "task",
                    "identifier": ridentifier,
                },
                status_code=status.HTTP_403_FORBIDDEN,
            )

    return {"identifiers": [task["task_identifier"] for task in tasks]}


@router.delete(
    "/task/{identifier}",
    summary="Delete ewoks task",
    response_model=common_models.ResourceInfo,
    response_description="Deleted ewoks task",
    status_code=200,
    responses={
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to read task",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_404_NOT_FOUND: {
            "description": "Task not found",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def delete_task(
    identifier: Annotated[
        str,
        Path(
            title="Task identifier",
            description="Unique identifier in the task database",
        ),
    ],
    settings: EwoksSettingsType,
) -> Dict[str, str]:
    try:
        json_backend.delete_resource(settings.resource_directory / "tasks", identifier)
    except PermissionError:
        return JSONResponse(
            {
                "message": f"No permission to delete task '{identifier}'.",
                "type": "task",
                "identifier": identifier,
            },
            status_code=status.HTTP_403_FORBIDDEN,
        )
    except FileNotFoundError:
        return JSONResponse(
            {
                "message": f"Task '{identifier}' is not found.",
                "type": "task",
                "identifier": identifier,
            },
            status_code=status.HTTP_404_NOT_FOUND,
        )
    return {"identifier": identifier, "type": "task"}



from typing import List, Optional, Dict
from pydantic import BaseModel
from pydantic import Field


class EwoksWorkflow(BaseModel):
    graph: Optional[Dict] = Field(title="Workflow attributes", default=None)
    nodes: Optional[List[Dict]] = Field(title="Node attributes", default=None)
    links: Optional[List[Dict]] = Field(title="Link attributes", default=None)


class EwoksWorkflowDescription(BaseModel):
    id: Optional[str] = Field(
        title="Workflow identifier unique to the server", default=None
    )
    label: Optional[str] = Field(
        title="Workflow label for human consumption", default=None
    )
    category: Optional[str] = Field(title="Workflow category", default=None)
    keywords: Optional[Dict] = Field(title="Workflow search keywords", default=None)
    input_schema: Optional[Dict] = Field(
        title="Workflow execute input schema for the frontend", default=None
    )
    ui_schema: Optional[Dict] = Field(
        title="Workflow execute UI schema for the frontend", default=None
    )


class EwoksWorkflowIdentifiers(BaseModel):
    identifiers: List[str] = Field(title="Workflow identifiers")


class EwoksWorkflowDescriptions(BaseModel):
    items: List[EwoksWorkflowDescription] = Field(title="Workflow descriptions")



from .router import v1_0_0_router as _v1_0_0_router
from .router import v1_1_0_router as _v1_1_0_router
from .socketio import create_socketio_app as _create_socketio_app

routers = {(1, 0, 0): _v1_0_0_router, (1, 1, 0): _v1_1_0_router}

app_creators = {(1, 0, 0): _create_socketio_app, (1, 1, 0): _create_socketio_app}



from typing import List
from pydantic import BaseModel
from pydantic import Field


class EwoksIcon(BaseModel):
    data_url: str = Field(title="Icon data url")


class EwoksIconIdentifiers(BaseModel):
    identifiers: List[str] = Field(title="Icon identifiers")



import asyncio
from typing import List, Any, Dict

from ..app.routes.execution import socketio


class SocketIOTestClient:
    def __init__(self):
        self._manager = socketio._MANAGER
        self._events = list()
        self._org_emit = self._manager._emit

    def __enter__(self) -> "SocketIOTestClient":
        self._manager._emit = self._emit
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.disconnect()
        self._manager._emit = self._org_emit

    def connect(self) -> None:
        return _run_coroutine(self._manager.connect(None, None, None))

    def disconnect(self) -> None:
        return _run_coroutine(self._manager.disconnect(None))

    def get_events(self) -> List[Dict]:
        events = self._events
        self._events = self._events[len(events) :]
        return events

    def is_running(self) -> bool:
        return _run_coroutine(self._manager.is_running())

    def _emit(self, loop, event) -> None:
        self._events.append(event)


def _run_coroutine(coroutine) -> Any:
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(coroutine)



import time

import pytest
from ewokscore.tests.examples.graphs import get_graph

from .api_versions import ROOT_ALL_VERSIONS, ROOT_V1_1_0


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_execute_with_celery(celery_exec_client, root):
    _test_execute(root, *celery_exec_client)


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_execute_without_celery(local_exec_client, root):
    _test_execute(root, *local_exec_client)


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_new_client_new_events(local_exec_client, root):
    client, sclient = local_exec_client
    _test_execute(root, client, sclient)
    sclient.disconnect()
    sclient.connect()
    time.sleep(1)
    assert not sclient.get_events()


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_execute_options(rest_client, mocked_local_submit, root):
    workflow = {
        "graph": {
            "id": "myworkflow",
            "label": "label1",
            "category": "cat1",
            "execute_arguments": {
                "engine": "ppf",
                "_slurm_spawn_arguments": {
                    "parameters": {"time_limit": 10, "partition": "nice"},
                    "pre_script": "module load ewoks",
                },
            },
            "worker_options": {"queue": "id00"},
        },
        "nodes": [{"id": "task1"}],
    }
    response = rest_client.post(f"{root}/workflows", json=workflow)
    data = response.json()
    assert response.status_code == 200, data

    # Check that the backend uses execute_arguments and worker_options
    # from the workflow definition
    response = rest_client.post(f"{root}/execute/myworkflow")
    expected_submit_arguments = {
        "args": (),
        "kwargs": {
            "args": (workflow,),
            "kwargs": {
                "engine": "ppf",
                "_slurm_spawn_arguments": {
                    "parameters": {"time_limit": 10, "partition": "nice"},
                    "pre_script": "module load ewoks",
                },
            },
            "queue": "id00",
        },
    }
    assert mocked_local_submit == expected_submit_arguments

    # Check that the backend merges execute_arguments and worker_options
    # from the client
    data = {
        "execute_arguments": {
            "engine": "ppf",
            "_slurm_spawn_arguments": {
                "parameters": {"time_limit": 20, "partition": "nice"},
                "pre_script": "module load ewoks",
            },
        },
        "worker_options": {"queue": "id00", "time_limit": 30},
    }

    response = rest_client.post(f"{root}/execute/myworkflow", json=data)
    expected_submit_arguments = {
        "args": (),
        "kwargs": {
            "args": (workflow,),
            "kwargs": {
                "engine": "ppf",
                "_slurm_spawn_arguments": {
                    "parameters": {"time_limit": 20, "partition": "nice"},
                    "pre_script": "module load ewoks",
                },
            },
            "queue": "id00",
            "time_limit": 30,
        },
    }
    assert mocked_local_submit == expected_submit_arguments


def _test_execute(root, client, sclient):
    graph_name, expected = upload_graph(root, client)
    response = client.post(f"{root}/execute/{graph_name}")
    assert response.status_code == 200, response.json()

    n = 2 * (len(expected) + 2)
    events = get_events(sclient, n)
    _assert_events(response, events, expected)
    return n


def upload_graph(root, client):
    graph_name = "acyclic1"
    graph, expected = get_graph(graph_name)
    response = client.post(f"{root}/workflows", json=graph)
    assert response.status_code == 200, response.json()
    return graph_name, expected


def get_events(sclient, nevents, timeout=10):
    t0 = time.time()
    events = list()
    while True:
        events.extend(sclient.get_events())
        if len(events) == nevents:
            break
        time.sleep(0.1)
        if time.time() - t0 > timeout:
            raise TimeoutError(f"Received {len(events)} instead of {nevents}")
    return events


def _assert_events(response, events, expected):
    n = 2 * (len(expected) + 2)
    assert len(events) == n

    job_id = response.json()["job_id"]
    for event in events:
        assert event["job_id"] == job_id
        if event["node_id"]:
            assert event["node_id"] in expected


@pytest.mark.parametrize("root", ROOT_V1_1_0)
def test_get_workers_with_celery(celery_exec_client, root):
    rest_client, _ = celery_exec_client
    response = rest_client.get(f"{root}/execution/workers")
    assert response.status_code == 200, response.json()
    assert len(response.json()["workers"]) == 1


@pytest.mark.parametrize("root", ROOT_V1_1_0)
def test_get_workers_without_celery(rest_client, root):
    response = rest_client.get(f"{root}/execution/workers")
    assert response.status_code == 200, response.json()
    assert response.json()["workers"] is None



from typing import Optional, List, Dict
from pydantic import BaseModel
from pydantic import Field


class EwoksTaskDescription(BaseModel):
    task_type: str = Field(title="One of the Ewoks task types")
    task_identifier: str = Field(title="Task identifier unique to the server")
    category: Optional[str] = Field(title="Task category", default=None)
    icon: Optional[str] = Field(
        title="Task icon identifier unique to the server", default=None
    )
    required_input_names: Optional[List[str]] = Field(
        title="Task required input names", default=None
    )
    optional_input_names: Optional[List[str]] = Field(
        title="Task optional input names", default=None
    )
    output_names: Optional[List[str]] = Field(title="Task output names", default=None)


class EwoksTaskIdentifiers(BaseModel):
    identifiers: List[str] = Field(title="Task identifiers")


class EwoksTaskDescriptions(BaseModel):
    items: List[EwoksTaskDescription] = Field(title="Task descriptions")


class EwoksTaskDiscovery(BaseModel):
    modules: Optional[List[str]] = Field(title="Ewoks task description", default=None)
    worker_options: Optional[Dict] = Field(title="Worker options", default=None)



from ewoksserver.app import routes

ROOT_ALL_VERSIONS = (
    f"{routes.BACKEND_PREFIX}",
    f"{routes.BACKEND_PREFIX}/v1",
    f"{routes.BACKEND_PREFIX}/v1_1_0",
    f"{routes.BACKEND_PREFIX}/v1_0_0",
)

ROOT_V1_1_0 = (
    f"{routes.BACKEND_PREFIX}",
    f"{routes.BACKEND_PREFIX}/v1",
    f"{routes.BACKEND_PREFIX}/v1_1_0",
)



import os
import sys
import importlib.util
from pathlib import Path
from typing import Optional, Dict
from typing_extensions import Annotated

from pydantic import Field
from pydantic import BaseModel
from fastapi import Depends

try:
    from ewoksweb.serverutils import get_test_config
except ImportError:
    get_test_config = None


class EwoksSettings(BaseModel):
    configured: bool = Field(
        default=False, title="Config or resource directory have been defined"
    )
    resource_directory: Path = Field(
        default=Path("."), title="Backend file resource directory"
    )
    ewoks: Optional[Dict] = Field(default=None, title="Ewoks configuration")
    celery: Optional[Dict] = Field(default=None, title="Celery configuration")
    without_events: bool = Field(default=False, title="Enable ewoks events")
    discover_tasks: bool = Field(default=False, title="Discover ewoks tasks on startup")


class AppSettings(BaseModel):
    no_older_versions: bool = (
        Field(default=False, title="Do not create end points for older API versions"),
    )


_APP_SETTINGS = None

_EWOKS_SETTINGS = None


def create_ewoks_settings(
    config: Optional[str] = None,
    directory: Optional[str] = None,
    without_events: bool = False,
    frontend_tests: bool = False,
    rediscover_tasks: bool = False,
) -> EwoksSettings:
    global _EWOKS_SETTINGS

    # Get configuration file
    filename = os.environ.get("EWOKSSERVER_SETTINGS")
    if config:
        filename = config
    if frontend_tests:
        if get_test_config is None:
            raise RuntimeError("ewoksweb is not installed")
        filename = get_test_config()

    # Extract settings from configuration file
    resource_directory = None
    ewoks = None
    celery = None
    if filename:
        spec = importlib.util.spec_from_file_location("ewoksserverconfig", filename)
        mod = importlib.util.module_from_spec(spec)
        sys.modules["ewoksserverconfig"] = mod
        spec.loader.exec_module(mod)
        resource_directory = getattr(mod, "RESOURCE_DIRECTORY", resource_directory)
        ewoks = getattr(mod, "EWOKS", ewoks)
        celery = getattr(mod, "CELERY", celery)

    # Overwrite resource directory
    if directory:
        resource_directory = directory
    if not resource_directory:
        resource_directory = "."

    configured = bool(filename) or bool(directory)

    _EWOKS_SETTINGS = EwoksSettings(
        configured=configured,
        resource_directory=resource_directory,
        ewoks=ewoks,
        celery=celery,
        without_events=without_events,
        discover_tasks=rediscover_tasks,
    )
    return _EWOKS_SETTINGS


def create_app_settings(no_older_versions: bool = False) -> None:
    global _APP_SETTINGS
    _APP_SETTINGS = AppSettings(no_older_versions=no_older_versions)
    return _APP_SETTINGS


def get_ewoks_settings():
    global _EWOKS_SETTINGS
    if _EWOKS_SETTINGS is not None:
        return _EWOKS_SETTINGS
    return create_ewoks_settings()


def get_app_settings():
    global _APP_SETTINGS
    if _APP_SETTINGS is not None:
        return _APP_SETTINGS
    return create_app_settings()


EwoksSettingsType = Annotated[EwoksSettings, Depends(get_ewoks_settings)]



import json
from typing import List, Dict, Optional
from typing_extensions import Annotated

from fastapi import APIRouter
from fastapi import Path
from fastapi import Body
from fastapi import Query
from fastapi.responses import JSONResponse
from fastapi import status


from ...backends import json_backend
from ...config import EwoksSettingsType
from ..common import models as common_models
from . import models
from . import descriptions

router = APIRouter()


@router.get(
    "/workflow/{identifier}",
    summary="Get ewoks workflow",
    response_model=models.EwoksWorkflow,
    response_model_exclude_none=True,
    response_description="Ewoks workflow",
    status_code=200,
    responses={
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to read workflow",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_404_NOT_FOUND: {
            "description": "Workflow not found",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def get_workflow(
    identifier: Annotated[
        str,
        Path(
            title="Workflow identifier",
            description="Unique identifier in the workflow database",
        ),
    ],
    settings: EwoksSettingsType,
) -> json_backend.ResourceContentType:
    try:
        return json_backend.load_resource(
            settings.resource_directory / "workflows", identifier
        )
    except PermissionError:
        return JSONResponse(
            {
                "message": f"No permission to read workflow '{identifier}'.",
                "type": "workflow",
                "identifier": identifier,
            },
            status_code=status.HTTP_403_FORBIDDEN,
        )
    except FileNotFoundError:
        return JSONResponse(
            {
                "message": f"Workflow '{identifier}' is not found.",
                "type": "workflow",
                "identifier": identifier,
            },
            status_code=status.HTTP_404_NOT_FOUND,
        )


@router.get(
    "/workflows",
    summary="Get all ewoks workflow identifiers",
    response_model=models.EwoksWorkflowIdentifiers,
    response_description="Ewoks workflow identifiers",
    status_code=200,
)
def get_workflow_identifiers(
    settings: EwoksSettingsType, kw: Annotated[Optional[List[str]], Query()] = None
) -> Dict[str, List[str]]:
    keywords = _compile_keywords(kw)
    root = settings.resource_directory / "workflows"
    if keywords:
        identifiers = [
            desc["id"]
            for desc in descriptions.workflow_descriptions(root, keywords=keywords)
        ]
    else:
        identifiers = list(json_backend.resource_identifiers(root))
    return {"identifiers": identifiers}


def _compile_keywords(kw: Optional[List[str]]) -> Optional[Dict]:
    if not kw:
        return
    keywords = dict()
    for s in kw:
        name, _, value = s.partition(":")
        try:
            value = json.loads(value)
        except json.decoder.JSONDecodeError:
            pass
        keywords[name] = value
    return keywords


@router.get(
    "/workflows/descriptions",
    summary="Get all ewoks workflow descriptions",
    response_model=models.EwoksWorkflowDescriptions,
    response_model_exclude_none=True,
    response_description="Ewoks workflow descriptions",
    status_code=200,
)
def get_workflows(
    settings: EwoksSettingsType, kw: Annotated[Optional[List[str]], Query()] = None
) -> Dict[str, List[Dict]]:
    keywords = _compile_keywords(kw)
    return {
        "items": list(
            descriptions.workflow_descriptions(
                settings.resource_directory / "workflows", keywords=keywords
            )
        )
    }


@router.put(
    "/workflow/{identifier}",
    summary="Update ewoks workflow",
    response_model=models.EwoksWorkflow,
    response_model_exclude_none=True,
    response_description="Ewoks workflow",
    status_code=200,
    responses={
        status.HTTP_400_BAD_REQUEST: {
            "description": "Wrong workflow identifier",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_404_NOT_FOUND: {
            "description": "Workflow not found",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to edit workflow",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def update_workflow(
    identifier: Annotated[
        str,
        Path(
            title="Workflow identifier",
            description="Unique identifier in the workflow database",
        ),
    ],
    workflow: Annotated[models.EwoksWorkflow, Body(title="Ewoks workflow")],
    settings: EwoksSettingsType,
) -> models.EwoksWorkflow:
    if workflow.graph:
        ridentifier = workflow.graph.get("id", identifier)
    else:
        ridentifier = None
    if identifier != ridentifier:
        return JSONResponse(
            {
                "message": f"Resource identifier '{identifier}' is not equal to '{ridentifier}'.",
                "type": "workflow",
                "identifier": identifier,
            },
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    exists = json_backend.resource_exists(
        settings.resource_directory / "workflows", identifier
    )
    if not exists:
        return JSONResponse(
            {
                "message": f"Workflow '{identifier}' is not found.",
                "type": "workflow",
                "identifier": identifier,
            },
            status_code=status.HTTP_404_NOT_FOUND,
        )

    try:
        json_backend.save_resource(
            settings.resource_directory / "workflows",
            identifier,
            workflow.model_dump(exclude_none=True),
        )
    except PermissionError:
        return JSONResponse(
            {
                "message": f"No permission to edit workflow '{identifier}'.",
                "type": "workflow",
                "identifier": identifier,
            },
            status_code=status.HTTP_403_FORBIDDEN,
        )

    return workflow


@router.post(
    "/workflows",
    summary="Create ewoks workflow",
    response_model=models.EwoksWorkflow,
    response_model_exclude_none=True,
    response_description="Ewoks workflow",
    status_code=200,
    responses={
        status.HTTP_400_BAD_REQUEST: {
            "description": "Workflow identifier missing",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_409_CONFLICT: {
            "description": "Workflow already exists",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to create workflow",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def create_workflow(
    workflow: Annotated[models.EwoksWorkflow, Body(title="Ewoks workflow")],
    settings: EwoksSettingsType,
) -> models.EwoksWorkflow:
    if workflow.graph:
        ridentifier = workflow.graph.get("id")
    else:
        ridentifier = None
    if ridentifier is None:
        return JSONResponse(
            {
                "message": "Workflow identifier missing",
                "type": "workflow",
            },
            status_code=status.HTTP_400_CONFLICT,
        )

    exists = json_backend.resource_exists(
        settings.resource_directory / "workflows", ridentifier
    )
    if exists:
        return JSONResponse(
            {
                "message": f"Workflow '{ridentifier}' already exists.",
                "type": "workflow",
                "identifier": ridentifier,
            },
            status_code=status.HTTP_409_CONFLICT,
        )

    try:
        json_backend.save_resource(
            settings.resource_directory / "workflows",
            ridentifier,
            workflow.model_dump(exclude_none=True),
        )
    except PermissionError:
        return JSONResponse(
            {
                "message": f"No permission to create workflow '{ridentifier}'.",
                "type": "workflow",
                "identifier": ridentifier,
            },
            status_code=status.HTTP_403_FORBIDDEN,
        )

    return workflow


@router.delete(
    "/workflow/{identifier}",
    summary="Delete ewoks workflow",
    response_model=common_models.ResourceInfo,
    response_description="Deleted ewoks workflow",
    status_code=200,
    responses={
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to read workflow",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_404_NOT_FOUND: {
            "description": "Workflow not found",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def delete_workflow(
    identifier: Annotated[
        str,
        Path(
            title="Workflow identifier",
            description="Unique identifier in the workflow database",
        ),
    ],
    settings: EwoksSettingsType,
) -> Dict[str, str]:
    try:
        json_backend.delete_resource(
            settings.resource_directory / "workflows", identifier
        )
    except PermissionError:
        return JSONResponse(
            {
                "message": f"No permission to delete workflow '{identifier}'.",
                "type": "workflow",
                "identifier": identifier,
            },
            status_code=status.HTTP_403_FORBIDDEN,
        )
    except FileNotFoundError:
        return JSONResponse(
            {
                "message": f"Workflow '{identifier}' is not found.",
                "type": "workflow",
                "identifier": identifier,
            },
            status_code=status.HTTP_404_NOT_FOUND,
        )
    return {"identifier": identifier, "type": "workflow"}



from pydantic import BaseModel
from pydantic import Field


class ResourceInfo(BaseModel):
    identifier: str = Field(title="Resource identifier")


class ResourceError(BaseModel):
    message: str = Field(title="Error message")
    type: str = Field(title="Resource type")


class ResourceIdentifierError(ResourceError):
    identifier: str = Field(title="Resource identifier")



import logging
from pprint import pformat

from fastapi import FastAPI

from .config import get_app_settings
from .cors import enable_cors
from .lifespan import fastapi_lifespan
from . import routes
from .routes import backend
from .routes import frontend
from .routes import tasks
from .routes import workflows
from .routes import icons
from .routes import execution

logger = logging.getLogger(__name__)


def create_app() -> FastAPI:
    """Create the main API instance"""
    settings = get_app_settings()

    backend.assert_route_versions(
        tasks.routers,
        workflows.routers,
        icons.routers,
        execution.routers,
        execution.app_creators,
    )
    all_routes = (
        backend.get_routes("tasks", tasks.routers),
        backend.get_routes("workflows", workflows.routers),
        backend.get_routes("icons", icons.routers),
        backend.get_routes("execution", execution.routers),
        backend.get_routes("execution", execution.app_creators, suffix="socket.io"),
    )
    version_tags = backend.extract_version_tags(all_routes)
    major, minor, patch = backend.extract_latest_version(all_routes)

    tags_metadata = [
        {"name": "tasks", "description": "Ewoks workflow tasks"},
        {"name": "workflows", "description": "Ewoks workflows"},
        {"name": "icons", "description": "Ewoks workflow icons"},
        {"name": "execution", "description": "Ewoks workflow execution"},
        *(
            {"name": tag, "description": f"Ewoks workflows API {tag}"}
            for tag in version_tags
        ),
    ]

    app = FastAPI(
        title="ewoks",
        summary="Edit and execute ewoks workflows",
        version=f"{major}.{minor}.{patch}",
        contact={
            "name": "ESRF",
            "url": "https://gitlab.esrf.fr/workflow/ewoks/ewoksserver/issues",
        },
        license_info={
            "name": "MIT",
            "identifier": "MIT",
        },
        openapi_tags=tags_metadata,
        lifespan=fastapi_lifespan,
        openapi_url=f"{routes.BACKEND_PREFIX}/openapi.json",
        docs_url=f"{routes.BACKEND_PREFIX}/docs",
        swagger_ui_oauth2_redirect_url=f"{routes.BACKEND_PREFIX}/docs/oauth2-redirect",
        redoc_url=f"{routes.BACKEND_PREFIX}/redoc",
    )

    enable_cors(app)

    backend.add_routes(
        app,
        all_routes,
        no_older_versions=settings.no_older_versions,
    )

    frontend.add_frontend(app)  # Needs to come last for some reason

    logger.debug(f"Routes \n {pformat(app.routes)}")

    return app



"""Start ewoks server from the command line (supports configuration)

..code: bash

    ewoks-server --reload
"""

import sys
import logging
from typing import Optional, List

import click
from uvicorn.main import main as uvicorn_cmd

from .app.config import create_ewoks_settings
from .app.config import create_app_settings


uvicorn_cmd = click.option(
    "--config",
    type=str,
    default=None,
    help="Path to the config python script (equivalent to the environment variable 'EWOKSSERVER_SETTINGS')",
)(uvicorn_cmd)

uvicorn_cmd = click.option(
    "--dir",
    type=str,
    default=None,
    help="Root directory for resources (e.g. workflows, tasks, icons descriptions)",
)(uvicorn_cmd)

uvicorn_cmd = click.option(
    "--without-events",
    is_flag=True,
    help="Disable Socket.IO app for event stream",
)(uvicorn_cmd)

uvicorn_cmd = click.option(
    "--frontend-tests",
    is_flag=True,
    help="Load frontend test configuration",
)(uvicorn_cmd)

uvicorn_cmd = click.option(
    "--rediscover-tasks",
    is_flag=True,
    help="Run task discovery on start up",
)(uvicorn_cmd)

uvicorn_cmd = click.option(
    "--no-older-versions",
    is_flag=True,
    help="Do not provide end-points for older versions of the Ewoks API",
)(uvicorn_cmd)


def ewoks_main(
    config: Optional[str] = None,  # ewoks parameter
    dir: Optional[str] = None,  # ewoks parameter
    without_events: bool = False,  # ewoks parameter
    frontend_tests: bool = False,  # ewoks parameter
    rediscover_tasks: bool = False,  # ewoks parameter
    no_older_versions: bool = False,  # app parameter
    log_level: Optional[str] = None,  # uvicorn parameter
    **kw
):
    if not log_level:
        log_level = "info"
    if log_level:
        level = logging.getLevelName(log_level.upper())
        logging.basicConfig(
            level=level, format="%(levelname)8s(BACKEND %(asctime)s): %(message)s"
        )
    create_app_settings(no_older_versions=no_older_versions)
    create_ewoks_settings(
        config=config,
        directory=dir,
        without_events=without_events,
        frontend_tests=frontend_tests,
        rediscover_tasks=rediscover_tasks,
    )
    return uvicorn_main(log_level=log_level, **kw)


uvicorn_main = uvicorn_cmd.callback
uvicorn_cmd.callback = ewoks_main


def main(argv: Optional[List[str]] = None) -> None:
    """Exposes the uvicorn CLI with a default APP factory"""
    if argv is None:
        argv = sys.argv[1:]
    if "--factory" not in argv:
        argv += ["--factory", "ewoksserver.app:create_app"]
    uvicorn_cmd(argv)


if __name__ == "__main__":
    sys.exit(main())



import os
from collections import Counter
from datetime import datetime

import pytest
from ewokscore.tests.examples.graphs import get_graph

from .test_execute import upload_graph
from .test_execute import get_events
from .api_versions import ROOT_ALL_VERSIONS


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_get_execution_events(local_exec_client, root):
    client, sclient = local_exec_client

    graph_name, expected = upload_graph(root, client)
    nevents = 0
    nevents_per_exec = 2 * (len(expected) + 2)

    # Test no events (nothing has been executed)
    response = client.get(f"{root}/execution/events")
    assert response.status_code == 200
    data = response.json()
    assert data == {"jobs": list()}

    # Execute workflow
    response = client.post(f"{root}/execute/{graph_name}")
    data = response.json()
    assert response.status_code == 200, data
    job_id1 = data["job_id"]
    nevents += nevents_per_exec

    # Wait until all events have been received over the Socket.IO connection
    events1 = get_events(sclient, nevents)

    # Query should return the same a what was recieved over the Socket.IO connection
    response = client.get(f"{root}/execution/events")
    assert response.status_code == 200
    events = response.json()["jobs"]
    assert len(events) == 1
    assert events[0] == events1

    response = client.get(f"{root}/execution/events?job_id={job_id1}")
    assert response.status_code == 200
    events = response.json()["jobs"]
    assert len(events) == 1
    assert events[0] == events1

    response = client.get(f"{root}/execution/events?context=job")
    assert response.status_code == 200
    events = response.json()["jobs"]
    assert len(events) == 1
    assert len(events[0]) == 2

    dtmid = datetime.now().astimezone()

    # Execute workflow
    response = client.post(f"{root}/execute/{graph_name}")
    data = response.json()
    assert response.status_code == 200, data
    job_id2 = data["job_id"]
    nevents += nevents_per_exec

    # Wait until all events have been received over the Socket.IO connection
    events2 = get_events(sclient, nevents_per_exec)

    response = client.get(f"{root}/execution/events")
    assert response.status_code == 200
    events = response.json()["jobs"]
    assert len(events) == 2
    assert events[0] == events1
    assert events[1] == events2

    response = client.get(f"{root}/execution/events?job_id={job_id2}")
    assert response.status_code == 200
    events = response.json()["jobs"]
    assert len(events) == 1
    assert events[0] == events2

    response = client.get(f"{root}/execution/events?context=job")
    assert response.status_code == 200
    events = response.json()["jobs"]
    assert len(events) == 2
    assert len(events[0]) == 2
    assert len(events[1]) == 2

    if os.name == "nt":
        return  # TODO: time filtering fails on windows

    # Test time Query
    midtime = dtmid.isoformat().replace("+", "%2b")
    response = client.get(f"{root}/execution/events?endtime={midtime}")
    assert response.status_code == 200
    events = response.json()["jobs"]
    assert len(events) == 1
    assert events[0] == events1

    response = client.get(f"{root}/execution/events?starttime={midtime}")
    assert response.status_code == 200
    events = response.json()["jobs"]
    assert len(events) == 1
    assert events[0] == events2


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_get_execution_events_parallel(local_exec_client, root):
    client, sclient = local_exec_client

    _, expected = get_graph("demo")
    nevents = 0
    nevents_per_exec = 2 * (len(expected) + 2)

    # Test no events (nothing has been executed)
    jobs = client.get(f"{root}/execution/events").json()["jobs"]
    assert not jobs

    # Execute workflows in parallel
    nruns = 3
    for _ in range(nruns):
        response = client.post(
            f"{root}/execute/demo",
            json={"execute_arguments": {"inputs": [{"name": "delay", "value": 0.1}]}},
        )
        data = response.json()
        assert response.status_code == 200, data
        nevents += nevents_per_exec

    # Get events from Socket.IO and REST API
    events_socketio = get_events(sclient, nevents)
    events_get = client.get(f"{root}/execution/events").json()["jobs"]

    # Check that we have all events from the Socket.IO connection
    nevents = Counter()
    for event in events_socketio:
        nevents[event["job_id"]] += 1
    assert set(nevents.values()) == {nevents_per_exec}

    # Check that we have all events from the REST API
    nevents = Counter()
    for job in events_get:
        for event in job:
            nevents[event["job_id"]] += 1
    assert set(nevents.values()) == {nevents_per_exec}

    # Check whether that events from the REST API are properly grouped per job
    assert len(events_get) == nruns
    for job in events_get:
        assert len({event["job_id"] for event in job}) == 1



import logging
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from starlette.responses import Response
from starlette.types import Scope

try:
    from ewoksweb.serverutils import get_static_root
except ImportError:
    get_static_root = None

from . import FRONTEND_PREFIX


logger = logging.getLogger(__name__)


class _StaticFiles(StaticFiles):
    async def get_response(self, path: str, scope: Scope) -> Response:
        """This gets called for every path that is not mounted or route that is not included."""
        if path in ("edit", "monitor"):
            path = ""
        return await super().get_response(path, scope)


def add_frontend(app: FastAPI) -> None:
    if get_static_root is None:
        logger.info("No frontend available to serve (pip install ewoksweb)")
    else:
        files = _StaticFiles(directory=get_static_root(), html=True)
        app.mount(f"{FRONTEND_PREFIX}", app=files, name="ewoksweb")



import logging
import asyncio
import threading
from datetime import datetime
from typing import Optional
from concurrent.futures import ThreadPoolExecutor

import socketio
from socketio.exceptions import ConnectionRefusedError

from . import events
from ...config import EwoksSettings

logger = logging.getLogger(__name__)


class EwoksEventManager:
    """Asynchronous manager of a Socket.IO application."""

    def __init__(self) -> None:
        # Disable Socket.IO CORS since it is managed by the CORS middleware
        # https://github.com/encode/starlette/issues/1309#issuecomment-953930195
        self._sio = socketio.AsyncServer(async_mode="asgi", cors_allowed_origins=[])
        self._app = socketio.ASGIApp(self._sio, socketio_path="")

        self._sio.on("connect")(self.connect)
        self._sio.on("disconnect")(self.disconnect)

        self._ewoks_settings = None
        self._stop_event = threading.Event()
        self._fetch_events_future: Optional[asyncio.Future] = None
        self._counter = 0
        self._executor = ThreadPoolExecutor(max_workers=1)

    def configure(self, ewoks_settings: EwoksSettings) -> None:
        self._ewoks_settings = ewoks_settings

    async def connect(self, *_) -> None:
        if self._ewoks_settings.without_events:
            raise ConnectionRefusedError("Socket.IO has been disabled")
        self._counter += 1
        await self._start()

    async def disconnect(self, *_) -> None:
        if self._ewoks_settings.without_events:
            return
        self._counter = max(self._counter - 1, 0)
        if self._counter == 0:
            await self._stop(timeout=3)

    async def is_running(self) -> bool:
        return await self._is_running(self._fetch_events_future)

    @staticmethod
    async def _is_running(future: Optional[asyncio.Future] = None) -> bool:
        return future is not None and not future.done()

    async def _start(self) -> None:
        if await self.is_running():
            return

        self._stop_event.clear()
        loop = asyncio.get_running_loop()
        self._fetch_events_future = loop.run_in_executor(
            self._executor, self._fetch_events_main, loop
        )

    async def _stop(self, timeout: Optional[float] = None) -> None:
        future = self._fetch_events_future
        if not await self._is_running(future):
            return
        self._stop_event.set()
        await asyncio.wait_for(future, timeout=timeout)

    def _fetch_events_main(self, loop) -> None:
        try:
            with events.reader_context(self._ewoks_settings) as reader:
                if reader is None:
                    raise RuntimeError("Ewoks event handlers not configured")
                starttime = datetime.now().astimezone()
                for event in reader.wait_events(
                    starttime=starttime, stop_event=self._stop_event
                ):
                    if self._stop_event.is_set():
                        break
                    self._emit(loop, event)
        except Exception as e:
            # TODO: client needs to recieve an error
            logger.exception(str(e))
            raise
        finally:
            self._fetch_events_future = None

    def _emit(self, loop, event) -> None:
        coroutine = self._sio.emit("Executing", event)
        future = asyncio.run_coroutine_threadsafe(coroutine, loop)
        future.result()


def create_socketio_app() -> socketio.ASGIApp:
    """Create the ASGI Socket.IO application when needed"""
    global _MANAGER
    if _MANAGER is not None:
        return _MANAGER._app

    _MANAGER = EwoksEventManager()
    return _MANAGER._app


def configure_socketio(ewoks_settings: EwoksSettings) -> None:
    global _MANAGER
    _MANAGER.configure(ewoks_settings)


_MANAGER = None



import os
import sys
import json
import argparse
from .app import create_app


def save(argv=None):
    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description="Save Redoc specs for EwoksServer")
    parser.add_argument(
        "filename",
        help="JSON filename in which to save the specs",
    )
    args = parser.parse_args(argv[1:])

    app = create_app()
    directory = os.path.dirname(args.filename)
    if directory:
        os.makedirs(directory, exist_ok=True)
    with open(args.filename, "w") as fd:
        json.dump(app.openapi(), fd)


if __name__ == "__main__":
    sys.exit(save())



from typing import Dict, List, Union, Optional
from collections import OrderedDict
from typing_extensions import Annotated

from fastapi import APIRouter
from fastapi import Body
from fastapi import Path
from fastapi import Depends
from fastapi import status
from fastapi import HTTPException
from fastapi.responses import JSONResponse

from ewoksutils import event_utils
from ewoksjob.client import get_workers, submit
from ewoksjob.client.local import submit as submit_local

from ...backends import json_backend
from ...config import EwoksSettingsType
from ..common import models as common_models
from . import models
from . import events

v1_0_0_router = APIRouter()
v1_1_0_router = APIRouter()


@v1_0_0_router.post(
    "/execute/{identifier}",
    summary="Execute workflow",
    response_model=models.EwoksJobInfo,
    response_description="Workflow execution job description",
    status_code=200,
    responses={
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to read workflow",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_404_NOT_FOUND: {
            "description": "Workflow not found",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def execute_workflow(
    settings: EwoksSettingsType,
    identifier: Annotated[
        str,
        Path(
            title="Task identifier",
            description="Unique identifier in the task database",
        ),
    ],
    options: Annotated[
        Optional[models.EwoksExecuteOptions], Body(title="Ewoks execute options")
    ] = None,
) -> Dict[str, Union[int, str]]:
    try:
        graph = json_backend.load_resource(
            settings.resource_directory / "workflows", identifier
        )
    except PermissionError:
        return JSONResponse(
            {
                "message": f"No permission to read workflow '{identifier}'.",
                "type": "workflow",
                "identifier": identifier,
            },
            status_code=status.HTTP_403_FORBIDDEN,
        )
    except FileNotFoundError:
        return JSONResponse(
            {
                "message": f"Workflow '{identifier}' is not found.",
                "type": "workflow",
                "identifier": identifier,
            },
            status_code=status.HTTP_404_NOT_FOUND,
        )

    if options is None:
        execute_arguments = None
        worker_options = None
    else:
        execute_arguments = options.execute_arguments
        worker_options = options.worker_options
    execute_arguments = json_backend.merge_mappings(
        graph["graph"].get("execute_arguments"), execute_arguments
    )
    submit_kwargs = json_backend.merge_mappings(
        graph["graph"].get("worker_options"), worker_options
    )

    # Workflow execution: position arguments
    submit_kwargs["args"] = (graph,)
    # Workflow execution: named arguments
    submit_kwargs["kwargs"] = execute_arguments

    ewoks_config = settings.ewoks
    if ewoks_config:
        execinfo = execute_arguments.setdefault("execinfo", dict())
        handlers = execinfo.setdefault("handlers", list())
        for handler in ewoks_config.get("handlers", list()):
            if handler not in handlers:
                handlers.append(handler)

    if settings.celery is None:
        future = submit_local(**submit_kwargs)
    else:
        future = submit(**submit_kwargs)
    return {"job_id": future.task_id}


@v1_0_0_router.get(
    "/execution/events",
    summary="Get workflow events",
    response_model=models.EwoksEventList,
    response_description="Workflow execution jobs grouped per job ID",
    status_code=200,
    responses={
        status.HTTP_503_SERVICE_UNAVAILABLE: {
            "description": "Server not configured for ewoks events",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def execute_events(
    settings: EwoksSettingsType,
    filters: Annotated[
        models.EwoksEventFilter, Depends(models.EwoksEventFilter)
    ],  # pydantic model to parse query parameters
) -> Dict[str, List[List[Dict]]]:
    jobs = OrderedDict()
    with events.reader_context(settings) as reader:
        if reader is None:
            raise HTTPException(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
                detail="Server not configured for ewoks events",
            )
        for event in reader.get_events(**filters.model_dump(exclude_none=True)):
            job_id = event["job_id"]
            if job_id not in jobs:
                jobs[job_id] = list()
            if "engine" in event_utils.FIELD_TYPES:
                event["binding"] = event.pop("engine")
            jobs[job_id].append(event)
    return {"jobs": list(jobs.values())}


@v1_1_0_router.get(
    "/execution/workers",
    summary="Get workers",
    response_model=models.EwoksWorkerList,
    response_description="List of available workers",
    status_code=200,
)
def workers(settings: EwoksSettingsType) -> Dict[str, Optional[List[str]]]:
    if settings.celery is None:
        return {"workers": None}

    return {"workers": get_workers()}


v1_1_0_router.include_router(v1_0_0_router)



import json
import logging
from pathlib import Path
from typing import Iterator, Union, Optional, Mapping, Dict


ResourceIdentifierType = str
ResourceUrlType = Path
ResourceContentType = Dict


_logger = logging.getLogger(__name__)


def root_url(root_url: Union[str, Path, None], category: str) -> ResourceUrlType:
    if not root_url:
        root_url = Path(".")
    elif isinstance(root_url, str):
        root_url = Path(root_url)
    return root_url / category


def resource_identifiers(root: ResourceUrlType) -> Iterator[ResourceIdentifierType]:
    for url in _resource_urls(root):
        yield _url_to_identifier(url)


def resources(root: ResourceUrlType) -> Iterator[ResourceContentType]:
    for url in _resource_urls(root):
        yield _load_url(url)


def resource_exists(root: ResourceUrlType, identifier: ResourceIdentifierType) -> bool:
    return _identifier_to_url(root, identifier).exists()


def _resource_urls(root: ResourceUrlType) -> Iterator[ResourceUrlType]:
    if not root.exists():
        return
    for url in root.iterdir():
        if _is_resource(url):
            yield url


def _is_resource(url: ResourceUrlType) -> bool:
    return url.is_file() and url.name.endswith(".json")


def save_resource(
    root: ResourceUrlType,
    identifier: ResourceIdentifierType,
    resource: ResourceContentType,
):
    url = _identifier_to_url(root, identifier)
    _save_url(url, resource)


def load_resource(
    root: ResourceUrlType, identifier: ResourceIdentifierType
) -> ResourceContentType:
    url = _identifier_to_url(root, identifier)
    return _load_url(url)


def delete_resource(root: ResourceUrlType, identifier: ResourceIdentifierType) -> None:
    url = _identifier_to_url(root, identifier)
    _delete_url(url)


def _identifier_to_url(root: ResourceUrlType, identifier: ResourceIdentifierType):
    return root / (identifier + ".json")


def _url_to_identifier(url: ResourceUrlType) -> ResourceIdentifierType:
    return url.stem


def _save_url(url: ResourceUrlType, resource: ResourceContentType):
    _logger.debug("Save file '%s'", url)
    url.parent.mkdir(parents=True, exist_ok=True)
    with open(url, "w") as f:
        json.dump(resource, f, indent=2)


def _load_url(url: ResourceUrlType) -> ResourceContentType:
    try:
        with open(url, "r") as f:
            return json.load(f)
    except FileNotFoundError:
        _logger.error(f"'{url}' not found")
        raise


def _delete_url(url: ResourceUrlType) -> ResourceContentType:
    _logger.debug("Delete file '%s'", url)
    url.unlink()


def merge_mappings(d1: Optional[Mapping], d2: Optional[Mapping]) -> dict:
    """Same as `{**d1, **d2}` but then recursive"""
    if d1 is None:
        merged = dict()
    else:
        merged = dict(d1)
    if not d2:
        return merged
    for key, value2 in d2.items():
        value1 = merged.get(key)
        if isinstance(value1, Mapping) and isinstance(value2, Mapping):
            value2 = merge_mappings(value1, value2)
        merged[key] = value2
    return merged



from pathlib import Path
from typing import List
from functools import lru_cache
from collections import namedtuple

import pytest
from fastapi.testclient import TestClient

from ewokscore import events
from ewoksjob.tests.conftest import celery_config  # noqa F401
from ewoksjob.tests.conftest import celery_includes  # noqa F401

from .. import app as newserver
from ..app import config as serverconfig
from ..app.backends.binary_backend import _load_url
from ..resources import DEFAULT_ROOT

from .data import resource_filenames
from .socketio_test import SocketIOTestClient


@pytest.fixture
def rest_client(tmpdir):
    """Client to the REST server (no execution)."""
    app = newserver.create_app()

    @lru_cache()
    def get_ewoks_settings_for_tests():
        return serverconfig.EwoksSettings(
            configured=True, resource_directory=str(tmpdir)
        )

    app.dependency_overrides[serverconfig.get_ewoks_settings] = (
        get_ewoks_settings_for_tests
    )

    with TestClient(app) as client:
        yield client


@pytest.fixture()
def ewoks_handlers(tmpdir):
    uri = f"file:{tmpdir / 'ewoks_events.db'}"
    yield [
        {
            "class": "ewokscore.events.handlers.Sqlite3EwoksEventHandler",
            "arguments": [{"name": "uri", "value": uri}],
        }
    ]
    events.cleanup()


@pytest.fixture
def local_exec_client(tmpdir, ewoks_handlers):
    """Client to the REST server and Socket.IO (execution with process pool)."""
    app = newserver.create_app()

    def get_settings_override():
        return serverconfig.EwoksSettings(
            configured=True,
            resource_directory=str(tmpdir),
            ewoks={"handlers": ewoks_handlers},
        )

    app.dependency_overrides[serverconfig.get_ewoks_settings] = get_settings_override

    with TestClient(app) as client:
        with SocketIOTestClient() as sclient:
            yield client, sclient


@pytest.fixture
def celery_exec_client(tmpdir, celery_session_worker, ewoks_handlers):
    """Client to the REST server and Socket.IO (execution with celery)."""
    app = newserver.create_app()

    def get_settings_override():
        return serverconfig.EwoksSettings(
            configured=True,
            resource_directory=str(tmpdir),
            celery=dict(),
            ewoks={"handlers": ewoks_handlers},
        )

    app.dependency_overrides[serverconfig.get_ewoks_settings] = get_settings_override

    with TestClient(app) as client:
        with SocketIOTestClient() as sclient:
            yield client, sclient


@pytest.fixture
def png_icons():
    filenames = resource_filenames()
    return [_load_url(filename) for filename in filenames if filename.endswith(".png")]


@pytest.fixture
def svg_icons():
    filenames = resource_filenames()
    return [_load_url(filename) for filename in filenames if filename.endswith(".svg")]


@pytest.fixture(scope="session")
def default_icon_identifiers() -> List[Path]:
    return [
        url.name
        for url in (DEFAULT_ROOT / "icons").iterdir()
        if not url.name.startswith("__")
    ]


@pytest.fixture(scope="session")
def default_workflow_identifiers() -> List[Path]:
    return [
        url.stem
        for url in (DEFAULT_ROOT / "workflows").iterdir()
        if url.suffix == ".json"
    ]


@pytest.fixture(scope="session")
def default_task_identifiers() -> List[Path]:
    return [
        url.stem for url in (DEFAULT_ROOT / "tasks").iterdir() if url.suffix == ".json"
    ]


@pytest.fixture
def mocked_local_submit(mocker) -> str:
    submit_local_mock = mocker.patch(
        "ewoksserver.app.routes.execution.router.submit_local"
    )

    MockFuture = namedtuple("Future", ["task_id"])

    arguments = dict()
    task_id = 0

    def mocked_submit(*args, **kwargs):
        nonlocal task_id
        arguments["args"] = args
        arguments["kwargs"] = kwargs
        task_id += 1
        return MockFuture(task_id=task_id)

    submit_local_mock.side_effect = mocked_submit
    return arguments



import logging
from contextlib import contextmanager
from typing import Generator, Optional

from ewoksjob.events.readers import EwoksEventReader
from ewoksjob.events.readers import instantiate_reader

from ...config import EwoksSettingsType

logger = logging.getLogger(__name__)


@contextmanager
def reader_context(
    ewoks_settings: EwoksSettingsType,
) -> Generator[Optional[EwoksEventReader], None, None]:
    r = _reader(ewoks_settings)
    try:
        yield r
    finally:
        if r is not None:
            r.close()


def _reader(ewoks_settings: EwoksSettingsType) -> Optional[EwoksEventReader]:
    cfg = ewoks_settings.ewoks
    if cfg:
        handlers = cfg.get("handlers", list())
    else:
        handlers = list()
    argmap = {"uri": "url"}
    for name in ("Redis", "Sqlite3", None):
        for handler in handlers:
            if name is None or name in handler["class"]:
                arguments = handler.get("arguments", list())
                arguments = {
                    argmap.get(arg["name"], arg["name"]): arg["value"]
                    for arg in arguments
                }
                return instantiate_reader(**arguments)

    logger.warning("Configure ewoks event handlers")
    return None



import os
import shutil
import logging
from pprint import pformat
from typing import Generator
from contextlib import contextmanager
from contextlib import asynccontextmanager

from fastapi import FastAPI
from ewoksjob.client.local import pool_context
from celery import current_app as current_celery_app

from .backends import json_backend
from .. import resources
from . import config
from .routes.execution import socketio
from .routes.tasks.discovery import discover_tasks


logger = logging.getLogger(__name__)


@asynccontextmanager
async def fastapi_lifespan(app: FastAPI) -> Generator[None, None, None]:
    get_ewoks_settings = app.dependency_overrides.get(
        config.get_ewoks_settings, config.get_ewoks_settings
    )
    ewoks_settings = get_ewoks_settings()
    _configure_socketio(ewoks_settings)
    _copy_default_resources(ewoks_settings)
    _enable_execution_events(ewoks_settings)
    with _enable_execution(ewoks_settings):
        _rediscover_tasks(ewoks_settings)
        _print_ewoks_settings(ewoks_settings)
        yield


def _configure_socketio(app_settings: config.EwoksSettings) -> None:
    socketio.configure_socketio(app_settings)


def _copy_default_resources(ewoks_settings: config.EwoksSettings) -> None:
    """Copy the default resources (tasks, workflows and icon) from the
    python package to the resource directory."""
    for resource, resource_ext in {
        "tasks": [".json"],
        "icons": [".png", ".svg"],
        "workflows": [".json"],
    }.items():
        root_url = json_backend.root_url(ewoks_settings.resource_directory, resource)
        os.makedirs(root_url, exist_ok=True)
        for filename in os.listdir(resources.DEFAULT_ROOT / resource):
            _, ext = os.path.splitext(filename)
            if ext not in resource_ext:
                continue

            src = resources.DEFAULT_ROOT / resource / filename
            if not os.path.isfile(src):
                continue

            dest = root_url / filename
            if not os.path.exists(dest):
                shutil.copy(src, dest)


def _rediscover_tasks(ewoks_settings: config.EwoksSettings) -> None:
    if not ewoks_settings.discover_tasks:
        return
    tasks = discover_tasks(ewoks_settings)
    root_url = json_backend.root_url(ewoks_settings.resource_directory, "tasks")
    for resource in tasks:
        json_backend.save_resource(root_url, resource["task_identifier"], resource)


def _enable_execution_events(ewoks_settings: config.EwoksSettings) -> None:
    """Set default ewoks event handler when nothing has been configured"""
    if ewoks_settings.configured:
        return
    if ewoks_settings.ewoks is None:
        ewoks_settings.ewoks = dict()
    if not ewoks_settings.ewoks.get("handlers"):
        ewoks_settings.ewoks["handlers"] = [
            {
                "class": "ewokscore.events.handlers.Sqlite3EwoksEventHandler",
                "arguments": [
                    {
                        "name": "uri",
                        "value": "file:ewoks_events.db",
                    }
                ],
            }
        ]


@contextmanager
def _enable_execution(
    ewoks_settings: config.EwoksSettings,
) -> Generator[None, None, None]:
    """Ensure workflows can be executed"""
    if ewoks_settings.celery is None:
        with pool_context():
            yield
    else:
        current_celery_app.conf.update(ewoks_settings.celery)
        yield


def _print_ewoks_settings(ewoks_settings: config.EwoksSettings) -> None:
    """Print summary of all Ewoks settings"""
    lines = list()
    resourcedir = ewoks_settings.resource_directory
    if not resourcedir:
        resourcedir = "."
    lines += ["", "", "RESOURCE DIRECTORY:", os.path.abspath(resourcedir)]

    adict = ewoks_settings.celery
    if adict is None:
        lines += ["", "CELERY:", "Not configured (local workflow execution)"]
    else:
        lines += ["", "CELERY:", pformat(adict)]

    adict = ewoks_settings.ewoks
    if adict is None:
        lines += ["", "EWOKS:", "Not configured (local workflow execution)"]
    else:
        lines += ["", "EWOKS:", pformat(adict)]

    lines += [""]

    logger.info("\n".join(lines))



from typing import List, Dict
from typing_extensions import Annotated

from fastapi import APIRouter
from fastapi import Path
from fastapi import Body
from fastapi.responses import JSONResponse
from fastapi import status


from ...backends import binary_backend
from ...config import EwoksSettingsType
from ..common import models as common_models
from . import models

router = APIRouter()


@router.get(
    "/icon/{identifier}",
    summary="Get ewoks icon",
    response_model=models.EwoksIcon,
    response_model_exclude_none=True,
    response_description="Ewoks icon",
    status_code=200,
    responses={
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to read icon",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_404_NOT_FOUND: {
            "description": "Icon not found",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def get_icon(
    identifier: Annotated[
        str,
        Path(
            title="Icon identifier",
            description="Unique identifier in the icon database",
        ),
    ],
    settings: EwoksSettingsType,
) -> binary_backend.ResourceContentType:
    try:
        return binary_backend.load_resource(
            settings.resource_directory / "icons", identifier
        )
    except PermissionError:
        return JSONResponse(
            {
                "message": f"No permission to read icon '{identifier}'.",
                "type": "icon",
                "identifier": identifier,
            },
            status_code=status.HTTP_403_FORBIDDEN,
        )
    except FileNotFoundError:
        return JSONResponse(
            {
                "message": f"Icon '{identifier}' is not found.",
                "type": "icon",
                "identifier": identifier,
            },
            status_code=status.HTTP_404_NOT_FOUND,
        )


@router.get(
    "/icons",
    summary="Get all ewoks icon identifiers",
    response_model=models.EwoksIconIdentifiers,
    response_description="Ewoks icon identifiers",
    status_code=200,
)
def get_icon_identifiers(settings: EwoksSettingsType) -> Dict[str, List[str]]:
    return {
        "identifiers": list(
            binary_backend.resource_identifiers(settings.resource_directory / "icons")
        )
    }


@router.put(
    "/icon/{identifier}",
    summary="Update ewoks icon",
    response_model=models.EwoksIcon,
    response_model_exclude_none=True,
    response_description="Ewoks icon",
    status_code=200,
    responses={
        status.HTTP_404_NOT_FOUND: {
            "description": "Icon not found",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to edit icon",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def update_icon(
    identifier: Annotated[
        str,
        Path(
            title="Icon identifier",
            description="Unique identifier in the icon database",
        ),
    ],
    icon: Annotated[models.EwoksIcon, Body(title="Ewoks icon")],
    settings: EwoksSettingsType,
) -> models.EwoksIcon:
    exists = binary_backend.resource_exists(
        settings.resource_directory / "icons", identifier
    )
    if not exists:
        return JSONResponse(
            {
                "message": f"Icon '{identifier}' is not found.",
                "type": "icon",
                "identifier": identifier,
            },
            status_code=status.HTTP_404_NOT_FOUND,
        )

    try:
        binary_backend.save_resource(
            settings.resource_directory / "icons", identifier, icon.model_dump()
        )
    except PermissionError:
        return JSONResponse(
            {
                "message": f"No permission to edit icon '{identifier}'.",
                "type": "icon",
                "identifier": identifier,
            },
            status_code=status.HTTP_403_FORBIDDEN,
        )

    return icon


@router.post(
    "/icon/{identifier}",
    summary="Create ewoks icon",
    response_model=models.EwoksIcon,
    response_model_exclude_none=True,
    response_description="Ewoks icon",
    status_code=200,
    responses={
        status.HTTP_409_CONFLICT: {
            "description": "Icon already exists",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to create icon",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def create_icon(
    identifier: Annotated[
        str,
        Path(
            title="Icon identifier",
            description="Unique identifier in the icon database",
        ),
    ],
    icon: Annotated[models.EwoksIcon, Body(title="Ewoks icon")],
    settings: EwoksSettingsType,
) -> models.EwoksIcon:
    exists = binary_backend.resource_exists(
        settings.resource_directory / "icons", identifier
    )
    if exists:
        return JSONResponse(
            {
                "message": f"Icon '{identifier}' already exists.",
                "type": "icon",
                "identifier": identifier,
            },
            status_code=status.HTTP_409_CONFLICT,
        )

    try:
        binary_backend.save_resource(
            settings.resource_directory / "icons", identifier, icon.model_dump()
        )
    except PermissionError:
        return JSONResponse(
            {
                "message": f"No permission to create icon '{identifier}'.",
                "type": "icon",
                "identifier": identifier,
            },
            status_code=status.HTTP_403_FORBIDDEN,
        )

    return icon


@router.delete(
    "/icon/{identifier}",
    summary="Delete ewoks icon",
    response_model=common_models.ResourceInfo,
    response_description="Deleted ewoks icon identifier",
    status_code=200,
    responses={
        status.HTTP_403_FORBIDDEN: {
            "description": "No permission to read icon",
            "model": common_models.ResourceIdentifierError,
        },
        status.HTTP_404_NOT_FOUND: {
            "description": "Icon not found",
            "model": common_models.ResourceIdentifierError,
        },
    },
)
def delete_icon(
    identifier: Annotated[
        str,
        Path(
            title="Icon identifier",
            description="Unique identifier in the icon database",
        ),
    ],
    settings: EwoksSettingsType,
) -> Dict[str, str]:
    try:
        binary_backend.delete_resource(
            settings.resource_directory / "icons", identifier
        )
    except PermissionError:
        return JSONResponse(
            {
                "message": f"No permission to delete icon '{identifier}'.",
                "type": "icon",
                "identifier": identifier,
            },
            status_code=status.HTTP_403_FORBIDDEN,
        )
    except FileNotFoundError:
        return JSONResponse(
            {
                "message": f"Icon '{identifier}' is not found.",
                "type": "icon",
                "identifier": identifier,
            },
            status_code=status.HTTP_404_NOT_FOUND,
        )
    return {"identifier": identifier, "type": "icon"}



import json
import pytest
from .api_versions import ROOT_ALL_VERSIONS


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_single_task(rest_client, root):
    identifier = "myproject.tasks.Dummy"

    response = rest_client.get(f"{root}/task/{identifier}")
    data = response.json()
    expected = {
        "message": "Task 'myproject.tasks.Dummy' is not found.",
        "type": "task",
        "identifier": "myproject.tasks.Dummy",
    }
    assert response.status_code == 404
    assert data == expected

    task1a = {
        "task_identifier": identifier,
        "task_type": "class",
        "required_input_names": ["a"],
    }
    response = rest_client.post(f"{root}/tasks", json=task1a)
    data = response.json()
    assert response.status_code == 200, data
    expected = {
        "required_input_names": ["a"],
        "task_identifier": "myproject.tasks.Dummy",
        "task_type": "class",
    }
    assert data == expected

    response = rest_client.get(f"{root}/task/{identifier}")
    data = response.json()
    assert response.status_code == 200, data
    assert data == task1a

    task1b = {
        "task_identifier": identifier,
        "task_type": "class",
        "required_input_names": ["a", "b"],
    }
    response = rest_client.put(f"{root}/task/{identifier}", json=task1b)
    data = response.json()
    assert response.status_code == 200, data
    expected = {
        "required_input_names": ["a", "b"],
        "task_identifier": "myproject.tasks.Dummy",
        "task_type": "class",
    }
    assert data == expected

    response = rest_client.get(f"{root}/task/{identifier}")
    data = response.json()
    assert response.status_code == 200, data
    assert data == task1b

    response = rest_client.delete(f"{root}/task/{identifier}")
    data = response.json()
    assert response.status_code == 200
    assert data == {"identifier": identifier}

    response = rest_client.delete(f"{root}/task/{identifier}")
    data = response.json()
    assert response.status_code == 404
    assert data == {
        "identifier": identifier,
        "message": f"Task '{identifier}' is not found.",
        "type": "task",
    }

    response = rest_client.get(f"{root}/task/{identifier}")
    data = response.json()
    assert response.status_code == 404
    expected = {
        "identifier": identifier,
        "message": f"Task '{identifier}' is not found.",
        "type": "task",
    }
    assert data == expected


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_multiple_tasks(rest_client, default_task_identifiers, root):
    response = rest_client.get(f"{root}/tasks")
    data = response.json()
    assert response.status_code == 200
    assert sorted(data["identifiers"]) == sorted(default_task_identifiers)

    task1a = {
        "task_identifier": "myproject.tasks.Dummy1",
        "task_type": "class",
        "required_input_names": ["a"],
    }
    task1b = {
        "task_identifier": "myproject.tasks.Dummy1",
        "task_type": "class",
        "required_input_names": ["a", "b"],
    }
    task2 = {
        "task_identifier": "myproject.tasks.Dummy2",
        "task_type": "class",
        "required_input_names": ["a", "b"],
    }

    response = rest_client.post(f"{root}/tasks", json=task1a)
    data = response.json()
    assert response.status_code == 200, data
    assert data == task1a

    response = rest_client.post(f"{root}/tasks", json=task1b)
    data = response.json()
    assert response.status_code == 409, data
    expected = {
        "identifier": "myproject.tasks.Dummy1",
        "message": "Task 'myproject.tasks.Dummy1' already exists.",
        "type": "task",
    }
    assert data == expected

    response = rest_client.post(f"{root}/tasks", json=task2)
    data = response.json()
    assert response.status_code == 200, data
    assert data == task2

    response = rest_client.get(f"{root}/tasks")
    data = response.json()
    assert response.status_code == 200, data
    expected = default_task_identifiers + [
        "myproject.tasks.Dummy1",
        "myproject.tasks.Dummy2",
    ]
    assert sorted(data["identifiers"]) == sorted(expected)


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_discover_tasks(rest_client, default_task_identifiers, root):
    response = rest_client.get(f"{root}/tasks")
    data = response.json()
    assert response.status_code == 200
    assert sorted(data["identifiers"]) == sorted(default_task_identifiers)

    module = "ewoksserver.tests.dummy_tasks"

    response = rest_client.post(f"{root}/tasks/discover", json={"modules": [module]})
    data = response.json()
    assert response.status_code == 200, data
    expected = [
        "ewoksserver.tests.dummy_tasks.MyTask1",
        "ewoksserver.tests.dummy_tasks.MyTask2",
    ]
    assert sorted(data["identifiers"]) == sorted(expected)

    response = rest_client.get(f"{root}/tasks")
    data = response.json()
    assert response.status_code == 200
    expected = default_task_identifiers + [
        "ewoksserver.tests.dummy_tasks.MyTask1",
        "ewoksserver.tests.dummy_tasks.MyTask2",
    ]
    assert sorted(data["identifiers"]) == sorted(expected)

    response = rest_client.post(f"{root}/tasks/discover", json={"modules": [module]})
    data = response.json()
    assert response.status_code == 200, data

    response = rest_client.post(
        f"{root}/tasks/discover", json={"modules": ["not_a_module"]}
    )
    data = response.json()
    assert response.status_code == 404, data
    assert "No module named" in data["message"]

    response = rest_client.post(f"{root}/tasks/discover")
    data = response.json()
    assert response.status_code == 200, data
    assert data["identifiers"]


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_task_descriptions(rest_client, default_task_identifiers, root):
    response = rest_client.get(f"{root}/tasks/descriptions")
    data = response.json()
    assert response.status_code == 200
    default_descriptions = [
        desc
        for desc in data["items"]
        if desc["task_identifier"] in default_task_identifiers
    ]
    assert data == {"items": default_descriptions}

    module = "ewoksserver.tests.dummy_tasks"

    response = rest_client.post(f"{root}/tasks/discover", json={"modules": [module]})
    data1 = response.json()
    assert response.status_code == 200, data1

    response = rest_client.get(f"{root}/tasks/descriptions")
    data2 = response.json()["items"]
    data2 = [
        r["task_identifier"] for r in data2 if r["task_identifier"].startswith(module)
    ]
    assert response.status_code == 200
    assert sorted(data1["identifiers"]) == sorted(data2)


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_malformed_task(rest_client, root, tmpdir):
    malformed_task_id = "myproject.tasks.Malformed"
    task_malformed = {"task_identifier": malformed_task_id}
    normal_task_id = "myproject.tasks.Normal"
    task_normal = {
        "task_identifier": normal_task_id,
        "task_type": "class",
    }
    response = rest_client.post(f"{root}/tasks", json=task_malformed)
    assert response.status_code == 422

    response = rest_client.get(f"{root}/tasks/descriptions")
    data = response.json()
    existing_tasks = len(data["items"])

    tasks_dir = tmpdir / "tasks"
    malformed_task_file = tasks_dir / f"{malformed_task_id}.json"
    normal_task_file = tasks_dir / f"{normal_task_id}.json"

    with open(normal_task_file, "w") as f:
        json.dump(task_normal, f)
    with open(malformed_task_file, "w") as f:
        json.dump(task_malformed, f)

    response = rest_client.get(f"{root}/tasks/{malformed_task_id}")
    assert response.status_code == 404

    response = rest_client.get(f"{root}/tasks/descriptions")
    data = response.json()
    assert response.status_code == 200
    assert len(data["items"]) - existing_tasks == 1

    response = rest_client.get(f"{root}/tasks")
    data = response.json()
    assert response.status_code == 200
    assert len(data["identifiers"]) - existing_tasks == 1



import pytest
from .api_versions import ROOT_ALL_VERSIONS


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_single_workflow(rest_client, root):
    identifier = "myworkflow"

    response = rest_client.get(f"{root}/workflow/{identifier}")
    assert response.status_code == 404

    workflow1a = {"graph": {"id": identifier}, "nodes": [{"id": "task1"}]}
    response = rest_client.post(f"{root}/workflows", json=workflow1a)
    data = response.json()
    assert response.status_code == 200, data
    assert data == workflow1a

    response = rest_client.get(f"{root}/workflow/{identifier}")
    data = response.json()
    assert response.status_code == 200, data
    assert data == workflow1a

    workflow1b = {"graph": {"id": identifier}, "nodes": [{"id": "task2"}]}
    response = rest_client.put(f"{root}/workflow/{identifier}", json=workflow1b)
    data = response.json()
    assert response.status_code == 200, data
    assert data == workflow1b

    response = rest_client.get(f"{root}/workflow/{identifier}")
    data = response.json()
    assert response.status_code == 200, data
    assert data == workflow1b

    response = rest_client.delete(f"{root}/workflow/{identifier}")
    data = response.json()
    assert response.status_code == 200
    assert data == {"identifier": identifier}

    response = rest_client.delete(f"{root}/workflow/{identifier}")
    data = response.json()
    assert response.status_code == 404
    assert data["message"] == f"Workflow '{identifier}' is not found."

    response = rest_client.get(f"{root}/workflow/{identifier}")
    data = response.json()
    assert response.status_code == 404
    assert data["message"] == f"Workflow '{identifier}' is not found."


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_multiple_workflows(rest_client, default_workflow_identifiers, root):
    response = rest_client.get(f"{root}/workflows")
    data = response.json()
    assert response.status_code == 200
    assert data == {"identifiers": list(default_workflow_identifiers)}

    workflow1a = {"graph": {"id": "myworkflow1"}, "nodes": [{"id": "task1"}]}
    workflow1b = {"graph": {"id": "myworkflow1"}, "nodes": [{"id": "task2"}]}
    workflow2 = {"graph": {"id": "myworkflow2"}, "nodes": [{"id": "task1"}]}

    response = rest_client.post(f"{root}/workflows", json=workflow1a)
    data = response.json()
    assert response.status_code == 200, data

    response = rest_client.post(f"{root}/workflows", json=workflow1b)
    data = response.json()
    assert response.status_code == 409, data
    assert data["message"] == "Workflow 'myworkflow1' already exists."
    response = rest_client.post(f"{root}/workflows", json=workflow2)
    data = response.json()
    assert response.status_code == 200, data

    response = rest_client.get(f"{root}/workflows")
    data = response.json()
    assert response.status_code == 200
    expected = default_workflow_identifiers + ["myworkflow1", "myworkflow2"]
    assert sorted(data["identifiers"]) == sorted(expected)


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_workflow_descriptions(rest_client, default_workflow_identifiers, root):
    response = rest_client.get(f"{root}/workflows/descriptions")
    data = response.json()
    assert response.status_code == 200
    default_descriptions = [
        desc for desc in data["items"] if desc["id"] in default_workflow_identifiers
    ]
    assert data == {"items": default_descriptions}

    workflow1 = {
        "graph": {"id": "myworkflow1", "label": "label1", "category": "cat1"},
        "nodes": [{"id": "task1"}],
    }
    workflow2 = {"graph": {"id": "myworkflow2"}, "nodes": [{"id": "task1"}]}
    response = rest_client.post(f"{root}/workflows", json=workflow1)
    data = response.json()
    assert response.status_code == 200, data
    response = rest_client.post(f"{root}/workflows", json=workflow2)
    data = response.json()
    assert response.status_code == 200, data

    response = rest_client.get(f"{root}/workflows/descriptions")
    data = response.json()["items"]
    assert response.status_code == 200
    expected = default_descriptions + [
        {"id": "myworkflow1", "label": "label1", "category": "cat1"},
        {"id": "myworkflow2"},
    ]
    data = sorted(data, key=lambda x: x["id"])
    assert data == expected


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_workflow_description_keys(rest_client, default_workflow_identifiers, root):
    desc = {
        "id": "myworkflow1",
        "label": "label1",
        "category": "cat1",
        "keywords": {"tags": ["XRPD", "ID00"]},
        "input_schema": {"title": "Demo workflow"},
        "ui_schema": {"mx_pipeline_name": {"ui:widget": "checkboxes"}},
    }
    workflow1 = {
        "graph": {
            **desc,
            "custom1": 1,
            "custom2": {},
            "execute_arguments": {"engine": "ppf"},
            "worker_options": {"queue": "id00"},
        },
        "nodes": [{"id": "task1"}],
    }
    response = rest_client.post(f"{root}/workflows", json=workflow1)
    data = response.json()
    assert response.status_code == 200, data

    response = rest_client.get(
        f'{root}/workflows/descriptions?kw=tags:["XRPD", "ID00"]'
    )
    data = response.json()["items"]
    assert data == [desc], data


@pytest.mark.parametrize("root", ROOT_ALL_VERSIONS)
def test_workflow_keywords(rest_client, default_workflow_identifiers, root):
    for instrument_name in ("ID00", "ID99"):
        for scan_type in ("ct", "loopscan"):
            workflow = {
                "graph": {
                    "id": f"myworkflow_{instrument_name}_{scan_type}",
                    "label": "label1",
                    "category": "cat1",
                    "keywords": {
                        "instrument_name": instrument_name,
                        "scan_type": scan_type,
                    },
                },
                "nodes": [{"id": "task1"}],
            }
            response = rest_client.post(f"{root}/workflows", json=workflow)
            data = response.json()
            assert response.status_code == 200, data

    response = rest_client.get(f"{root}/workflows")
    data = response.json()["identifiers"]
    assert response.status_code == 200
    expected = default_workflow_identifiers + [
        "myworkflow_ID00_ct",
        "myworkflow_ID00_loopscan",
        "myworkflow_ID99_ct",
        "myworkflow_ID99_loopscan",
    ]
    assert sorted(data) == sorted(expected)

    response = rest_client.get(f"{root}/workflows?kw=instrument_name:ID00")
    data = response.json()["identifiers"]
    assert response.status_code == 200
    expected = ["myworkflow_ID00_ct", "myworkflow_ID00_loopscan"]
    assert sorted(data) == sorted(expected)

    response = rest_client.get(
        f"{root}/workflows?kw=instrument_name:ID00&kw=scan_type:ct"
    )
    data = response.json()["identifiers"]
    assert response.status_code == 200
    assert data == ["myworkflow_ID00_ct"]

    response = rest_client.get(f"{root}/workflows/descriptions")
    data = [res["id"] for res in response.json()["items"]]
    expected = default_workflow_identifiers + [
        "myworkflow_ID00_ct",
        "myworkflow_ID00_loopscan",
        "myworkflow_ID99_ct",
        "myworkflow_ID99_loopscan",
    ]
    assert sorted(data) == sorted(expected)

    response = rest_client.get(f"{root}/workflows/descriptions?kw=instrument_name:ID00")
    data = [res["id"] for res in response.json()["items"]]
    expected = ["myworkflow_ID00_ct", "myworkflow_ID00_loopscan"]
    assert sorted(data) == sorted(expected)

    response = rest_client.get(
        f"{root}/workflows/descriptions?kw=instrument_name:ID00&kw=scan_type:ct",
    )
    data = [res["id"] for res in response.json()["items"]]
    assert data == ["myworkflow_ID00_ct"]



from typing import Dict, Optional, Iterator

from ...backends import json_backend


_WORKFLOW_KEYWORDS = (
    "id",
    "label",
    "category",
    "keywords",
    "input_schema",
    "ui_schema",
)


def workflow_descriptions(
    root: json_backend.ResourceUrlType, keywords: Optional[Dict] = None
) -> Iterator[Dict]:
    for res in json_backend.resources(root):
        description = res["graph"]
        if not _include_resource(description.get("keywords", dict()), keywords):
            continue
        yield {
            key: value
            for key, value in description.items()
            if key in _WORKFLOW_KEYWORDS
        }


def _include_resource(res_keywords: dict, keywords: Optional[Dict] = None) -> bool:
    if keywords is None:
        return True
    return all(res_keywords.get(key) == value for key, value in keywords.items())



from typing import Optional, List, Union, Dict
from pydantic import BaseModel
from pydantic import Field


class EwoksExecuteOptions(BaseModel):
    execute_arguments: Optional[Dict] = Field(
        title="Workflow execution options", default=None
    )
    worker_options: Optional[Dict] = Field(title="Worker options", default=None)


class EwoksJobInfo(BaseModel):
    job_id: Union[str, int] = Field(title="Workflow execution job identifier")


class EwoksEvent(BaseModel):
    host_name: str = Field(title="Host where the job was executed")
    process_id: int = Field(title="Process ID where the job was executed")
    user_name: str = Field(title="User name under which the job was executed")
    job_id: str = Field(title="Workflow execution job identifier")
    binding: Optional[str] = Field(title="Workflow execution engine", default=None)
    context: str = Field(title="Event context (job, workflow, node)")
    workflow_id: Optional[str] = Field(title="Workflow identifier", default=None)
    node_id: Optional[str] = Field(title="Workflow node identifier", default=None)
    task_id: Optional[str] = Field(title="Workflow task identifier", default=None)
    type: str = Field(title="Event type (start, end, progress)")
    time: str = Field(title="Event context send time")
    error: Optional[bool] = Field(title="Workflow execution failed", default=None)
    error_message: Optional[str] = Field(
        title="Workflow execution error message", default=None
    )
    error_traceback: Optional[str] = Field(
        title="Workflow execution error traceback", default=None
    )
    progress: Optional[int] = Field(title="Task progress in percentage", default=None)
    task_uri: Optional[str] = Field(title="Workflow task output URI", default=None)
    input_uris: Optional[List[Dict]] = Field(
        title="Workflow task input URIs", default=None
    )
    output_uris: Optional[List[Dict]] = Field(
        title="Workflow task output URIs", default=None
    )


class EwoksEventFilter(BaseModel):
    user_name: Optional[str] = Field(
        title="User name under which the job was executed", default=None
    )
    job_id: Optional[str] = Field(
        title="Workflow execution job identifier", default=None
    )
    context: Optional[str] = Field(
        title="Event context (job, workflow, node)", default=None
    )
    workflow_id: Optional[str] = Field(title="Workflow identifier", default=None)
    node_id: Optional[str] = Field(title="Workflow node identifier", default=None)
    task_id: Optional[str] = Field(title="Workflow task identifier", default=None)
    type: Optional[str] = Field(title="Event type (start, end, progress)", default=None)
    starttime: Optional[str] = Field(title="Only events after this time", default=None)
    endtime: Optional[str] = Field(title="Only events before this time", default=None)
    error: Optional[bool] = Field(title="Workflow execution failed", default=None)


class EwoksEventList(BaseModel):
    jobs: List[List[EwoksEvent]] = Field(
        title="Workflow execution jobs grouped per job ID"
    )


class EwoksWorkerList(BaseModel):
    workers: Optional[List[str]] = Field(title="Available workers for execution")
